/*----------------------------------------------------------------------
  File    : matrix4.c
  Contents: matrix functions for special problems
  Author  : Christian Borgelt
  History : 2008.09.20 file created
            2011.02.04 function mat_negate() added (negate elements)
            2011.06.21 bug for non-square matrices in mat_match() fixed
            2013.08.13 adapted to preprocessor definition of DIMID
----------------------------------------------------------------------*/
#include <stdio.h>
#include <limits.h>
#include <float.h>
#include <assert.h>
#include "matrix.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Hungarian Method for Bipartite Matching
----------------------------------------------------------------------*/

double mat_match (MATRIX *mat, DIMID *map)
{                               /* --- bipartite matching */
  DIMID  row, col, cur, n;      /* loop variables, stack size */
  DIMID  *inv, *stk;            /* inverse map and row stack */
  double *r;                    /* to traverse the matrix rows */
  double t;                     /* buffer for a matrix element */
  double min;                   /* minimum element (of a row etc.) */
  double cost = 0;              /* total cost of matching */

  assert(mat                    /* check the function arguments */
  &&    (mat->colcnt >= mat->rowcnt));
  if (!map) map = mat->map;     /* get internal map if necessary */
  inv = mat->map +mat->rowcnt;  /* and the (internal) inverse map */
  for (row = mat->rowcnt; --row >= 0; ) {
    min = *(r = mat->els[row]); /* find the minimal element in row */
    for (col = mat->colcnt; --col >  0; )
      if (r[col] < min) min = r[col];
    for (col = mat->colcnt; --col >= 0; )
      r[col] -= min;            /* subtract minimum from the row */
    cost += min;                /* and collect the subtracted costs */
    map[row] = DIMID_MAX;       /* init. the row to column map */
  }                             /* (no matching found yet) */
  /* Since one element of each row must be chosen for the matching,  */
  /* the minimum of each row can be subtracted from the row elements */
  /* and added to the total costs. This does not change the solution,*/
  /* but ensures that all matrix elements are positive and that the  */
  /* trivial labeling of the rows and the columns consists of zeros. */
  if (mat->colcnt != mat->rowcnt)
    for (col = mat->colcnt; --col >= 0; )
      inv[col] = DIMID_MAX;     /* init. the column to row map */
  else {                        /* if matrix is square */
    for (col = mat->colcnt; --col >= 0; ) {
      min = mat->els[0][col];   /* find the minimal element in column */
      for (row = mat->rowcnt; --row >  0; )
        if ((t = mat->els[row][col]) < min) min = t;
      for (row = mat->rowcnt; --row >= 0; )
      mat->els[row][col] -= min;/* subtract minimum from the column */
      cost += min;              /* and collect the subtracted costs */
      inv[col] = DIMID_MAX;     /* init. the column to row map */
    }                           /* (no matching found yet) */
  }
  /* If the matrix is square, one element of each column must be   */
  /* chosen for the matching. Therefore the minimum of each column */
  /* can be subtracted from the column elements and added to the   */
  /* total costs. This does not change the solution, but reduces   */
  /* the size of the numbers that have to be handled. The trivial  */
  /* labeling of rows and columns is unaffected by this operation. */
  for (row = mat->rowcnt; --row >= 0; ) {
    r = mat->els[row];          /* traverse the matrix rows */
    for (col = mat->colcnt; --col >= 0; )
      if ((inv[col] >= DIMID_MAX) && (r[col] <= 0)) {
        map[row] = col; inv[col] = row; break; }
  }                             /* assign rows whenever possible */
  /* Vanishing matrix elements indicate edges of the equality graph */
  /* for the initial labeling. An initial, not necessarily complete */
  /* matching is selected from this equality graph. Note that each  */
  /* row has at least one incident edge in the equality graph.      */
  stk = (DIMID*)mat->buf;       /* init. the stack of tree leaves */
  for (cur = mat->rowcnt; --cur >= 0; ) {
    if (map[cur] < DIMID_MAX)   /* find the next unmatched row */
      continue;                 /* (skip all matched rows) */
    map[cur] |= DIMID_SGN;      /* mark the root of the alt. tree */
    stk[0] = cur; n = 1;        /* and push it onto the stack */
    while (1) {                 /* loop to extend the matching */
      row = stk[--n];           /* pop a row from the stack */
      r   = mat->els[row];      /* and then get this row */
      for (col = mat->colcnt; --col >= 0; ) {
        if (r[col] > 0)         /* traverse the columns to which */
          continue;             /* the row may be assigned */
        /* Only edges with a vanishing weight are eligible, because  */
        /* these are the edges of the equality graph of the labeling.*/
        /* Note that each row has at least one such incident edge.   */
        if (inv[col] >= DIMID_MAX)  /* if an assignment is possible, */
          break;                    /* abort the loop */
        if (inv[col] >= 0) {    /* if not yet in alternating tree */
          map[stk[n++] = inv[col]] |= DIMID_SGN;
          inv[col] = row | DIMID_SGN;
        }                       /* push the row onto the stack */
      }                         /* and link the column to the row */
      if (col >= 0) {           /* if augmenting altern. path found */
        while (map[row] != (DIMID_SGN|DIMID_MAX)) {
          n = map[row] & ~DIMID_SGN; /* follow the path to the root */
          map[row] = col;            /* and reassign the rows */
          row = inv[col = n] & ~DIMID_SGN;
        }                       /* (free a column for the root row) */
        map[cur] = col; break;  /* assign the current row (root) */
      }                         /* to the freed column and abort */
      /* After a reassignment, the matching has been extended    */
      /* by the row that forms the root of the alternating tree. */
      if (n > 0) continue;      /* check for an empty stack */
      /* If the stack is not empty, the alternating tree is not  */
      /* yet complete, so an extension of the tree is needed.    */
      /* Otherwise the equality graph has to be extended, i.e.   */
      /* the cost matrix has to be adapted to create more zeros. */
      min = INFINITY;           /* adapt the matrix elements */
      for (row = mat->rowcnt; --row >= 0; ) {
        if (map[row] >= 0) continue;
        r = mat->els[row];      /* traverse rows in alternating tree */
        for (col = mat->colcnt; --col >= 0; ) {
          if ((inv[col] >= 0) && (r[col] < min))
            min = r[col];       /* find the minimum cost for a */
        }                       /* marked row and unmarked column */
      }                         /* (best alternative assigment) */
      /* Since the current equality graph contains no augmenting path */
      /* for the current matching, some next best edge must be added. */
      assert(min > 0);          /* check for a positive minimum */
      cost += min;              /* collect the subtracted costs */
      for (row = mat->rowcnt; --row >= 0; ) {
        r   = mat->els[row];    /* traverse all matrix rows */
        col = mat->colcnt;      /* and all matrix columns */
        if (map[row] >= 0) {    /* if the row is unmarked, */
          while (--col >= 0)    /* add minimum to marked columns */
            if (inv[col] < 0) r[col] += min; }
        else {                  /* if row is marked (in altern. tree) */
          while (--col >= 0) {  /* subtract min. from unmarked cols. */
            if (inv[col] < 0) continue;
            if ((r[col] -= min) <= 0) stk[n++] = row;
          }                     /* enlarge the equality graph, i.e. */
        }                       /* increase the number of zeros */
      }                         /* in the cost matrix */
    }
    for (row = mat->rowcnt; --row >= 0; ) {
      col = map[row] &= ~DIMID_SGN;
      if (col < DIMID_MAX) inv[col] = row;
    }                           /* clear markers and restore/update */
  }                             /* the column to row mapping */
  return cost;                  /* return the total matching costs */
}  /* mat_match() */

/*--------------------------------------------------------------------*/

double mat_negate (MATRIX *mat, double zero)
{                               /* --- negate matrix to given zero */
  DIMID  row, col;              /* loop variables */
  double *r;                    /* to traverse the matrix rows */

  assert(mat);                  /* check the function arguments */
  if (isnan(zero) || isinf(zero)){ /* if no zero value is given */
    for (row = mat->rowcnt; --row >= 0; ) {
      r = mat->els[row];        /* traverse the matrix rows */
      for (col = mat->colcnt; --col >= 0; )
        if (r[col] > zero) zero = r[col];
    }                           /* traverse the matrix elements */
  }                             /* and find the maximum element */
  for (row = mat->rowcnt; --row >= 0; ) {
    r = mat->els[row];          /* traverse the matrix rows */
    for (col = mat->colcnt; --col >= 0; )
      r[col] = zero -r[col];    /* traverse the matrix elements */
  }                             /* and subtract them from maximum */
  return zero;                  /* return the used zero value */
}  /* mat_negate() */
